@@ -15,4 +15,20 @@ module FakeRedis
   def self.disable
     Redis::Connection.drivers.delete_if {|driver| Redis::Connection::Memory == driver }
   end
+
+  def self.disabling
+    return yield unless enabled?
+
+    disable
+    yield
+    enable
+  end
+
+  def self.enabling
+    return yield if enabled?
+
+    enable
+    yield
+    disable
+  end
 end
