@@ -5,6 +5,10 @@ module FakeRedis
       super(key, _floatify(val))
     end
 
+    def identical_scores?
+      values.uniq.size == 1
+    end
+
     # Increments the value of key by val
     def increment(key, val)
       self[key] += _floatify(val)
@@ -13,7 +17,7 @@ module FakeRedis
     def select_by_score min, max
       min = _floatify(min, true)
       max = _floatify(max, false)
-      reject {|_,v| v < min || v > max }
+      select {|_,v| v >= min && v <= max }
     end
 
     private
@@ -25,8 +29,10 @@ module FakeRedis
       elsif (( number = str.to_s.match(/^\((\d+)/i) ))
         number[1].to_i + (increment ? 1 : -1)
       else
-        Float str
+        Float str.to_s
       end
+    rescue ArgumentError
+      raise Redis::CommandError, "ERR value is not a valid float"
     end
 
   end
