@@ -9,16 +9,20 @@ require "fakeredis/sorted_set_store"
 require "fakeredis/transaction_commands"
 require "fakeredis/zset"
 require "fakeredis/bitop_command"
+require "fakeredis/geo_commands"
 require "fakeredis/version"
 
 class Redis
   module Connection
+    DEFAULT_REDIS_VERSION = '3.3.5'
+
     class Memory
       include Redis::Connection::CommandHelper
       include FakeRedis
       include SortMethod
       include TransactionCommands
       include BitopCommand
+      include GeoCommands
       include CommandExecutor
 
       attr_accessor :options
@@ -51,7 +55,7 @@ class Redis
       end
 
       def initialize(options = {})
-        self.options = options
+        self.options = self.options ? self.options.merge(options) : options
       end
 
       def database_id
@@ -93,9 +97,8 @@ class Redis
 
       def client(command, _options = {})
         case command
-        when :setname then true
+        when :setname then "OK"
         when :getname then nil
-        when :client then true
         else
           raise Redis::CommandError, "ERR unknown command '#{command}'"
         end
@@ -130,7 +133,7 @@ class Redis
 
       def info
         {
-          "redis_version" => "2.6.16",
+          "redis_version" => options[:version] || DEFAULT_REDIS_VERSION,
           "connected_clients" => "1",
           "connected_slaves" => "0",
           "used_memory" => "3187",
@@ -147,9 +150,13 @@ class Redis
 
       def save; end
 
-      def bgsave ; end
+      def bgsave; end
 
-      def bgrewriteaof ; end
+      def bgrewriteaof; end
+
+      def evalsha; end
+
+      def eval; end
 
       def move key, destination_id
         raise Redis::CommandError, "ERR source and destination objects are the same" if destination_id == database_id
@@ -214,6 +221,11 @@ class Redis
         data[key][start_index..end_index].unpack('B*')[0].count("1")
       end
 
+      def bitpos(key, bit, start_index = 0, end_index = -1)
+        value = data[key] || ""
+        value[0..end_index].unpack('B*')[0].index(bit.to_s, start_index * 8) || -1
+      end
+
       def getrange(key, start, ending)
         return unless data[key]
         data[key][start..ending]
@@ -493,9 +505,9 @@ class Redis
 
       def brpoplpush(key1, key2, opts={})
         data_type_check(key1, Array)
-        brpop(key1).tap do |elem|
-          lpush(key2, elem) unless elem.nil?
-        end
+        _key, elem = brpop(key1)
+        lpush(key2, elem) unless elem.nil?
+        elem
       end
 
       def lpop(key)
@@ -549,6 +561,8 @@ class Redis
 
       def srem(key, value)
         data_type_check(key, ::Set)
+        value = Array(value)
+        raise_argument_error('srem') if value.empty?
         return false unless data[key]
 
         if value.is_a?(Array)
@@ -575,7 +589,7 @@ class Redis
         data_type_check(key, ::Set)
         results = (count || 1).times.map do
           elem = srandmember(key)
-          srem(key, elem)
+          srem(key, elem) if elem
           elem
         end.compact
         count.nil? ? results.first : results
@@ -680,14 +694,11 @@ class Redis
       end
 
       def del(*keys)
-        keys = keys.flatten(1)
-        raise_argument_error('del') if keys.empty?
+        delete_keys(keys, 'del')
+      end
 
-        old_count = data.keys.size
-        keys.each do |key|
-          data.delete(key)
-        end
-        old_count - data.keys.size
+      def unlink(*keys)
+        delete_keys(keys, 'unlink')
       end
 
       def setnx(key, value)
@@ -798,7 +809,7 @@ class Redis
       end
 
       def hmget(key, *fields)
-        raise_argument_error('hmget')  if fields.empty?
+        raise_argument_error('hmget')  if fields.empty? || fields.flatten.empty?
 
         data_type_check(key, Hash)
         fields.flatten.map do |field|
@@ -817,6 +828,12 @@ class Redis
         data[key].size
       end
 
+      def hstrlen(key, field)
+        data_type_check(key, Hash)
+        return 0 if data[key].nil? || data[key][field].nil?
+        data[key][field].size
+      end
+
       def hvals(key)
         data_type_check(key, Hash)
         return [] unless data[key]
@@ -853,22 +870,14 @@ class Redis
 
       def sync ; end
 
-      def [](key)
-        get(key)
-      end
-
-      def []=(key, value)
-        set(key, value)
-      end
-
       def set(key, value, *array_options)
         option_nx = array_options.delete("NX")
         option_xx = array_options.delete("XX")
 
-        return false if option_nx && option_xx
+        return nil if option_nx && option_xx
 
-        return false if option_nx && exists(key)
-        return false if option_xx && !exists(key)
+        return nil if option_nx && exists(key)
+        return nil if option_xx && !exists(key)
 
         data[key] = value.to_s
 
@@ -899,6 +908,10 @@ class Redis
         "OK"
       end
 
+      def psetex(key, milliseconds, value)
+        setex(key, milliseconds / 1000.0, value)
+      end
+
       def setrange(key, offset, value)
         return unless data[key]
         s = data[key][offset,value.size]
@@ -1005,6 +1018,19 @@ class Redis
       end
 
       def zadd(key, *args)
+        option_xx = args.delete("XX")
+        option_nx = args.delete("NX")
+        option_ch = args.delete("CH")
+        option_incr = args.delete("INCR")
+
+        if option_xx && option_nx
+          raise_options_error("XX", "NX")
+        end
+
+        if option_incr && args.size > 2
+          raise_options_error("INCR")
+        end
+
         if !args.first.is_a?(Array)
           if args.size < 2
             raise_argument_error('zadd')
@@ -1020,18 +1046,39 @@ class Redis
         data_type_check(key, ZSet)
         data[key] ||= ZSet.new
 
-        if args.size == 2 && !(Array === args.first)
-          score, value = args
-          exists = !data[key].key?(value.to_s)
+        # Turn [1, 2, 3, 4] into [[1, 2], [3, 4]] unless it is already
+        args = args.each_slice(2).to_a unless args.first.is_a?(Array)
+
+        changed = 0
+        exists = args.map(&:last).count { |el| !hexists(key, el.to_s) }
+
+        args.each do |score, value|
+          if option_nx && hexists(key, value.to_s)
+            next
+          end
+
+          if option_xx && !hexists(key, value.to_s)
+            exists -= 1
+            next
+          end
+
+          if option_incr
+            data[key][value.to_s] ||= 0
+            return data[key].increment(value, score).to_s
+          end
+
+          if option_ch && data[key][value.to_s] != score
+            changed += 1
+          end
           data[key][value.to_s] = score
-        else
-          # Turn [1, 2, 3, 4] into [[1, 2], [3, 4]] unless it is already
-          args = args.each_slice(2).to_a unless args.first.is_a?(Array)
-          exists = args.map(&:last).map { |el| data[key].key?(el.to_s) }.count(false)
-          args.each { |s, v| data[key][v.to_s] = s }
         end
 
-        exists
+        if option_incr
+          changed = changed.zero? ? nil : changed
+          exists = exists.zero? ? nil : exists
+        end
+
+        option_ch ? changed : exists
       end
 
       def zrem(key, value)
@@ -1047,6 +1094,36 @@ class Redis
         response
       end
 
+      def zpopmax(key, count = nil)
+        data_type_check(key, ZSet)
+        return [] unless data[key]
+        sorted_members = sort_keys(data[key])
+        results = sorted_members.last(count || 1).reverse!
+        results.each do |member|
+          zrem(key, member.first)
+        end
+        count.nil? ? results.first : results.flatten
+      end
+
+      def zpopmin(key, count = nil)
+        data_type_check(key, ZSet)
+        return [] unless data[key]
+        sorted_members = sort_keys(data[key])
+        results = sorted_members.first(count || 1)
+        results.each do |member|
+          zrem(key, member.first)
+        end
+        count.nil? ? results.first : results.flatten
+      end
+
+      def bzpopmax(*args)
+        bzpop(:bzpopmax, args)
+      end
+
+      def bzpopmin(*args)
+        bzpop(:bzpopmin, args)
+      end
+
       def zcard(key)
         data_type_check(key, ZSet)
         data[key] ? data[key].size : 0
@@ -1055,7 +1132,13 @@ class Redis
       def zscore(key, value)
         data_type_check(key, ZSet)
         value = data[key] && data[key][value.to_s]
-        value && value.to_s
+        if value == Float::INFINITY
+          "inf"
+        elsif value == -Float::INFINITY
+          "-inf"
+        elsif value
+          value.to_s
+        end
       end
 
       def zcount(key, min, max)
@@ -1069,7 +1152,14 @@ class Redis
         data[key] ||= ZSet.new
         data[key][value.to_s] ||= 0
         data[key].increment(value.to_s, num)
-        data[key][value.to_s].to_s
+
+        if num =~ /^\+?inf/
+          "inf"
+        elsif num == "-inf"
+          "-inf"
+        else
+          data[key][value.to_s].to_s
+        end
       end
 
       def zrank(key, value)
@@ -1093,6 +1183,7 @@ class Redis
         results = sort_keys(data[key])
         # Select just the keys unless we want scores
         results = results.map(&:first) unless with_scores
+        start = [start, -results.size].max
         (results[start..stop] || []).flatten.map(&:to_s)
       end
 
@@ -1208,6 +1299,33 @@ class Redis
         data[out].size
       end
 
+      def pfadd(key, member)
+        data_type_check(key, Set)
+        data[key] ||= Set.new
+        previous_size = data[key].size
+        data[key] |= Array(member)
+        data[key].size != previous_size
+      end
+
+      def pfcount(*keys)
+        keys = keys.flatten
+        raise_argument_error("pfcount") if keys.empty?
+        keys.each { |key| data_type_check(key, Set) }
+        if keys.count == 1
+          (data[keys.first] || Set.new).size
+        else
+          union = keys.map { |key| data[key] }.compact.reduce(&:|)
+          union.size
+        end
+      end
+
+      def pfmerge(destination, *sources)
+        sources.each { |source| data_type_check(source, Set) }
+        union = sources.map { |source| data[source] || Set.new }.reduce(&:|)
+        data[destination] = union
+        "OK"
+      end
+
       def subscribe(*channels)
         raise_argument_error('subscribe') if channels.empty?()
 
@@ -1348,13 +1466,36 @@ class Redis
           raise Redis::CommandError, "ERR syntax error"
         end
 
+        def raise_options_error(*options)
+          if options.detect { |opt| opt.match(/incr/i) }
+            error_message = "ERR INCR option supports a single increment-element pair"
+          else
+            error_message = "ERR #{options.join(" and ")} options at the same time are not compatible"
+          end
+          raise Redis::CommandError, error_message
+        end
+
+        def raise_command_error(message)
+          raise Redis::CommandError, message
+        end
+
+        def delete_keys(keys, command)
+          keys = keys.flatten(1)
+          raise_argument_error(command) if keys.empty?
+
+          old_count = data.keys.size
+          keys.each do |key|
+            data.delete(key)
+          end
+          old_count - data.keys.size
+        end
+
         def remove_key_for_empty_collection(key)
           del(key) if data[key] && data[key].empty?
         end
 
         def data_type_check(key, klass)
           if data[key] && !data[key].is_a?(klass)
-            warn "Operation against a key holding the wrong kind of value: Expected #{klass} at #{key}."
             raise Redis::CommandError.new("WRONGTYPE Operation against a key holding the wrong kind of value")
           end
         end
@@ -1419,9 +1560,33 @@ class Redis
           end
         end
 
+        def bzpop(command, args)
+          timeout =
+            if args.last.is_a?(Hash)
+              args.pop[:timeout]
+            elsif args.last.respond_to?(:to_int)
+              args.pop.to_int
+            end
+
+          timeout ||= 0
+          single_pop_command = command.to_s[1..-1]
+          keys = args.flatten
+          keys.each do |key|
+            if data[key]
+              data_type_check(data[key], ZSet)
+              if data[key].size > 0
+                result = public_send(single_pop_command, key)
+                return result.unshift(key)
+              end
+            end
+          end
+          sleep(timeout.to_f)
+          nil
+        end
+
         def sort_keys(arr)
           # Sort by score, or if scores are equal, key alphanum
-          sorted_keys = arr.sort do |(k1, v1), (k2, v2)|
+          arr.sort do |(k1, v1), (k2, v2)|
             if v1 == v2
               k1 <=> k2
             else
