@@ -103,6 +103,34 @@ module FakeRedis
       expect(@client.exists("key")).to eq(false)
     end
 
+    it "should pop members with the highest scores from sorted set" do
+      @client.zadd("key", [1, "val1", 2, "val2", 3, "val3"])
+      expect(@client.zpopmax("key")).to eq(["val3", 3.0])
+      expect(@client.zpopmax("key", 3)).to eq([["val2", 2.0], ["val1", 1.0]])
+      expect(@client.zpopmax("nonexistent")).to eq(nil)
+    end
+
+    it "should pop members with the lowest scores from sorted set" do
+      @client.zadd("key", [1, "val1", 2, "val2", 3, "val3"])
+      expect(@client.zpopmin("key")).to eq(["val1", 1.0])
+      expect(@client.zpopmin("key", 3)).to eq([["val2", 2.0], ["val3", 3.0]])
+      expect(@client.zpopmin("nonexistent")).to eq(nil)
+    end
+
+    it "should pop members with the highest score from first sorted set that is non-empty" do
+      @client.zadd("key1", [1, "val1", 2, "val2"])
+      @client.zadd("key2", [3, "val3"])
+      expect(@client.bzpopmax("nonexistent", "key1", "key2", 0)).to eq(["key1", "val2", 2.0])
+      expect(@client.bzpopmax("nonexistent")).to eq(nil)
+    end
+
+    it "should pop members with the lowest score from first sorted set that is non-empty" do
+      @client.zadd("key1", [1, "val1", 2, "val2"])
+      @client.zadd("key2", [3, "val3"])
+      expect(@client.bzpopmin("nonexistent", "key1", "key2", 0)).to eq(["key1", "val1", 1.0])
+      expect(@client.bzpopmin("nonexistent")).to eq(nil)
+    end
+
     it "should get the number of members in a sorted set" do
       @client.zadd("key", 1, "val2")
       @client.zadd("key", 2, "val1")
@@ -147,6 +175,7 @@ module FakeRedis
 
       expect(@client.zrange("key", 0, -1)).to eq(["one", "two", "three"])
       expect(@client.zrange("key", 1, 2)).to eq(["two", "three"])
+      expect(@client.zrange("key", -50, -2)).to eq(["one", "two"])
       expect(@client.zrange("key", 0, -1, :withscores => true)).to eq([["one", 1], ["two", 2], ["three", 3]])
       expect(@client.zrange("key", 1, 2, :with_scores => true)).to eq([["two", 2], ["three", 3]])
     end
@@ -574,7 +603,7 @@ module FakeRedis
         expect(ranged).to eq %w(c b)
       end
 
-      it "should return empty array if key is not exist" do
+      it "should return empty array if key does not exist" do
         ranged = @client.zrevrangebylex "puppies", "+", "-"
         expect(ranged).to be_empty
       end
@@ -590,5 +619,154 @@ module FakeRedis
         expect(ranged).to eq %w(d c b)
       end
     end
+
+    describe "#zadd" do
+      context "with {incr: true}" do
+        before { @client.zadd("key", 1, "existing") }
+        it "should increment the element's score with the provided value" do
+          @client.zadd("key", 99, "existing", incr: true)
+          expect(@client.zscore("key", "existing")).to eq(100.0)
+
+          @client.zadd("key", 2, "new", incr: true)
+          expect(@client.zscore("key", "new")).to eq(2.0)
+        end
+
+        it "should error when trying to add multiple increment-element pairs" do
+          expect {
+            @client.zadd("key", [1, "member1", 2, "member2"], incr: true)
+          }.to raise_error(Redis::CommandError, "ERR INCR option supports a single increment-element pair")
+        end
+      end
+
+      context "with {nx: true}" do
+        before { @client.zadd("key", [1, "existing1", 2, "existing2"]) }
+        it "should add new elements but not update the scores of existing elements" do
+          @client.zadd("key", [101, "existing1", 3, "new"], nx: true)
+
+          expect(@client.zscore("key", "existing1")).to eq(1.0)
+          expect(@client.zscore("key", "new")).to eq(3.0)
+
+          @client.zadd("key", 102, "existing2", nx: true)
+          expect(@client.zscore("key", "existing2")).to eq(2.0)
+        end
+      end
+
+      context "with {xx: true}" do
+        before { @client.zadd("key", 1, "existing") }
+        it "should not add new elements" do
+          expect(@client.zadd("key", 1, "new1", xx: true)).to eq(false)
+          expect(@client.zscore("key", "new1")).to be_nil
+
+          expect(@client.zadd("key", [11, "existing", 2, "new2"], xx: true)).to eq(0)
+          expect(@client.zscore("key", "existing")).to eq(11.0)
+          expect(@client.zscore("key", "new2")).to be_nil
+        end
+      end
+
+      context "with {ch: true}" do
+        it "should return the number of new elements added plus the number of existing elements for which the score was updated" do
+          expect(@client.zadd("key", 1, "first", ch: true)).to eq(true)
+
+          expect(@client.zadd("key", [1, "first", 2, "second"], ch: true)).to eq(1.0)
+          expect(@client.zadd("key", [11, "first", 2, "second"], ch: true)).to eq(1.0)
+          expect(@client.zadd("key", [99, "first", 99, "second"], ch: true)).to eq(2.0)
+          expect(@client.zadd("key", [111, "first", 22, "second", 3, "third"], ch: true)).to eq(3.0)
+        end
+      end
+
+      context "with {nx: true, xx: true}" do
+        it "should error" do
+          expect{
+            @client.zadd("key", 1, "value", nx: true, xx: true)
+          }.to raise_error(Redis::CommandError, "ERR XX and NX options at the same time are not compatible")
+        end
+      end
+
+      context "with {nx: true, incr: true}" do
+        let(:options) { {nx: true, incr: true} }
+        it "should increment to the provided score only if the element is new and return the element's score" do
+          expect(@client.zadd("key", 1, "first", options)).to eq(1.0)
+          expect(@client.zscore("key", "first")).to eq(1.0)
+
+          expect(@client.zadd("key", 2, "second", options)).to eq(2.0)
+          expect(@client.zscore("key", "second")).to eq(2.0)
+
+          expect(@client.zadd("key", 99, "first", options)).to be_nil
+          expect(@client.zscore("key", "first")).to eq(1.0)
+        end
+      end
+
+      context "with {nx: true, ch: true}" do
+        let(:options) { {nx: true, ch: true} }
+        it "should add only new elements, not update existing elements, and return the number of added elements" do
+          expect(@client.zadd("key", 1, "first", options)).to eq(true)
+          expect(@client.zadd("key", 1, "first", options)).to eq(false)
+
+          # add two new elements
+          expect(@client.zadd("key", [99, "first", 2, "second", 3, "third"], options)).to eq(2)
+          expect(@client.zscore("key", "first")).to eq(1.0)
+        end
+      end
+
+      context "with {nx: true, incr: true, ch: true}" do
+        let(:options) { {nx: true, incr: true, ch: true} }
+
+        it "should add only new elements" do
+          expect(@client.zadd("key", 1, "first", options)).to eq(1.0)
+          expect(@client.zadd("key", 99, "first", options)).to be_nil
+          expect(@client.zscore("key", "first")).to eq(1.0)
+        end
+
+        # when INCR is present, return value is always the new score of member
+        it "should return the score of the new member" do
+          expect(@client.zadd("key", 2, "second", options)).to eq(2.0)
+        end
+
+        it "should return nil when the member already exists" do
+          @client.zadd("key", 1, "first")
+          expect(@client.zadd("key", 99, "first", options)).to be_nil
+        end
+      end
+
+      context "with {xx: true, incr: true}" do
+        let(:options) { {xx: true, incr: true} }
+        before { @client.zadd("key", 1, "existing") }
+
+        it "should return nil if the member does not already exist" do
+          expect(@client.zadd("key", 1, "new1", options)).to be_nil
+          expect(@client.zscore("key", "new1")).to be_nil
+        end
+
+        it "should increment only existing elements" do
+          expect(@client.zadd("key", [11, "existing"], options)).to eq(12.0)
+          expect(@client.zscore("key", "existing")).to eq(12.0)
+        end
+      end
+
+      context "with {xx: true, ch: true}" do
+        let(:options) { {xx: true, ch: true} }
+        it "should return the number of updated elements and not add new members" do
+          @client.zadd("key", 1, "first")
+
+          expect(@client.zadd("key", 99, "first", options)).to eq(true)
+          expect(@client.zadd("key", [100, "first", 2, "second"], options)).to eq(1.0)
+          expect(@client.zscore("key", "second")).to be_nil
+        end
+      end
+
+      context "with {xx: true, incr: true, ch: true}" do
+        let(:options) { {xx: true, incr: true, ch: true} }
+        before { @client.zadd("key", 1, "existing") }
+
+        # when INCR is present, return value is always the new score of member
+        it "should return the new score of the inserted member" do
+          expect(@client.zadd("key", 2, "existing", options)).to eq(3.0)
+        end
+
+        it "should increment only existing elements" do
+          expect(@client.zadd("key", 1, "new", options)).to be_nil
+        end
+      end
+    end
   end
 end
